from typing import Any

import numpy as np
import pandas as pd
from imblearn.pipeline import Pipeline
from sklearn.metrics import roc_curve, f1_score, auc
from sklearn.model_selection import (
    GridSearchCV,
    RepeatedStratifiedKFold,
    StratifiedKFold,
)

from racoons.models import classifiers, feature_selection_methods


def hyper_parameter_optimization(
    model: Pipeline, X: pd.DataFrame, y: pd.Series
) -> tuple[dict[Any, Any, Any], float, float]:
    """
    Optimize the hyperparameters of the given model using grid search cross-validation.

    Args:
        model (Pipeline): The machine learning model to be optimized.
        X (pd.DataFrame): The feature matrix.
        y (pd.Series): The target variable.

    Returns:
        Tuple[Dict[Any, Any, Any], float, float]: A tuple containing:
            - The best hyperparameters found during optimization.
            - Mean F1 score on the test sets for the best hyperparameters.
            - Standard deviation of F1 score on the test sets for the best hyperparameters.

    Raises:
        Any specific exceptions or errors.

    Example:
        >>> from sklearn.ensemble import RandomForestClassifier
        >>> best_params, mean_f1, std_f1 = hyper_parameter_optimization(
        ...     RandomForestClassifier(),
        ...     X_train,
        ...     y_train
        ... )

    Note:
        - This function performs hyperparameter optimization using grid search with cross-validation.
        - The search space for hyperparameters is defined by the get_param_grid() function.
        - Cross-validation is conducted using RepeatedStratifiedKFold with 5 splits and 2 repeats.
        - Scoring is based on both F1 score and accuracy, with F1 score used for refitting.
        - The function returns the best hyperparameters and associated mean and standard deviation of the F1 score.

    """
    params = get_param_grid(model)
    grid_cv = GridSearchCV(
        estimator=model,
        param_grid=params,
        cv=RepeatedStratifiedKFold(n_splits=5, n_repeats=2),
        scoring=["f1", "accuracy"],
        refit="f1",
        n_jobs=-1,
    )
    grid_cv.fit(X, y)
    return (
        grid_cv.best_params_,
        grid_cv.cv_results_["mean_test_f1"][grid_cv.best_index_],
        grid_cv.cv_results_["std_test_f1"][grid_cv.best_index_],
    )


def get_param_grid(model: Pipeline):
    """
    Get the hyperparameter grid for the specified model.

    The hyperparameter grid contains candidate values for hyperparameters to be used
    in hyperparameter optimization during model training.

    Args:
        model (Pipeline): The machine learning model pipeline.

    Returns:
        dict: A dictionary representing the hyperparameter grid.

    Example:
        >> from imblearn.pipeline import Pipeline
        >> from sklearn.ensemble import RandomForestClassifier
        >> from sklearn.feature_selection import SelectFromModel
        >> model = Pipeline(steps=[
        ...     ('preprocessor', ...),
        ...     ('sampling', ...),
        ...     ('feature_selection', SelectFromModel(RandomForestClassifier())),
        ...     ('estimator', RandomForestClassifier())
        ... ])
        >>> get_param_grid(model)
        {'feature_selection__estimator__C': [0.01, 0.1, 1, 10, 100],
         'feature_selection__estimator__solver': ['liblinear'],
         'estimator__n_estimators': [100, 500, 1000],
         'estimator__n_jobs': [-1]}

    Note:
        This function generates a hyperparameter grid based on the type of classifier in the pipeline.
        It is used for hyperparameter optimization using GridSearchCV.
    """
    param_grid = {}
    classifier = list(classifiers.keys())[
        list(classifiers.values()).index(model["estimator"])
    ]
    if "feature_selection" in model.named_steps.keys():
        feature_selection_method = list(feature_selection_methods.keys())[
            list(feature_selection_methods.values()).index(model["feature_selection"])
        ]
        if feature_selection_method == "lasso":
            param_grid["feature_selection__estimator__C"] = [0.01, 0.1, 1, 10, 100]
            param_grid["feature_selection__estimator__solver"] = ["liblinear"]
            # param_grid["feature_selection__estimator__n_jobs"] = [-1]
    if classifier == "logistic_regression":
        param_grid["estimator__penalty"] = ["l1", "l2"]
        param_grid["estimator__solver"] = ["liblinear"]
        param_grid["estimator__C"] = [0.3, 0.5, 0.8, 1, 10, 100]
        # param_grid["estimator__n_jobs"] = [-1]
    elif classifier == "random_forest":
        param_grid["estimator__n_estimators"] = [100, 500, 1000]
        param_grid["estimator__n_jobs"] = [-1]
    elif classifier == "ada_boost":
        param_grid["estimator__learning_rate"] = np.arange(0.1, 2.1, 0.1)
        param_grid["estimator__n_estimators"] = [10, 50, 100, 500, 1000]
    elif classifier == "gradient_boosting":
        param_grid["estimator__learning_rate"] = np.arange(0.1, 2.1, 0.1)
        param_grid["estimator__n_estimators"] = [10, 50, 100, 500, 1000]
    elif classifier == "decision_tree":
        param_grid["estimator__criterion"] = ["gini", "log_loss", "entropy"]
    elif classifier == "xgboost":
        param_grid["estimator__n_estimators"] = [10, 50, 100, 500, 1000]
    # elif classifier == "k_neighbors":
    #     param_grid["estimator__n_neighbors"] = [2, 5, 10, 50]
    #     param_grid["estimator__n_jobs"] = [-1]
    return param_grid


def cross_validate_model(model: Pipeline, X: pd.DataFrame, y: pd.Series) -> tuple:
    """
    Perform cross-validation for a given classification model.

    Args:
        model (Pipeline): The machine learning model pipeline.
        X (pd.DataFrame): The feature matrix.
        y (pd.Series): The target variable.

    Returns:
        tuple: A tuple containing lists of true positive rates, area under the ROC curve, F1 score and a dataframe of feature importances for each fold.

    Raises:
        None

    Example:
        >> from imblearn.pipeline import Pipeline
        >> from sklearn.ensemble import RandomForestClassifier
        >> from sklearn.model_selection import train_test_split
        >> X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
        >> model = Pipeline(steps=[
        ...     ('preprocessor', ...),
        ...     ('sampling', ...),
        ...     ('feature_selection', SelectFromModel(RandomForestClassifier())),
        ...     ('estimator', RandomForestClassifier())
        ... ])
        >> tprs, aucs, f1_scores = cross_validate_model(model, X_train, y_train)

    Note:
        This function uses StratifiedKFold with 10 folds for cross-validation, and it returns lists of true positive rates,
        area under the ROC curve, F1 scores and feature importances for each fold.
    """
    tprs = []
    aucs = []
    f1 = []
    feature_importance = pd.DataFrame()
    mean_fpr = np.linspace(0, 1, 100)
    cv = StratifiedKFold(10)
    for fold, (train, test) in enumerate(cv.split(X, y)):
        X_train = X.loc[train, :]
        X_test = X.loc[test, :]
        y_train = y[train]
        y_test = y[test]
        model.fit(X_train, y_train)
        y_pred = model.predict(X_test)
        fpr, tpr, _ = roc_curve(y_test, y_pred)
        f1.append(f1_score(y_test, y_pred))
        interp_tpr = np.interp(mean_fpr, fpr, tpr)
        interp_tpr[0] = 0.0
        tprs.append(interp_tpr)
        aucs.append(auc(fpr, tpr))
        feature_importance = pd.concat(
            [feature_importance, get_feature_importance(model)], axis=0
        )

    return tprs, aucs, f1, feature_importance


def metrics_from_cv_result(cv_result: tuple):
    """
    Computes results from the cv
    Args:
        cv_result (tuple): true positive rates, aucs, f1-scores

    Returns:
        dict: means and standard of the cross-validation metrics
    """
    tprs, aucs, f1, _ = cv_result
    mean_fpr = np.linspace(0, 1, 100)
    mean_tpr = np.mean(tprs, axis=0)
    mean_tpr[-1] = 1.0
    mean_auc = auc(mean_fpr, mean_tpr)
    std_auc = np.std(aucs)
    std_tpr = np.std(tprs, axis=0)
    mean_f1 = np.mean(f1)
    std_f1 = np.std(f1)
    metrics = {
        "mean_fpr": mean_fpr,
        "mean_tpr": mean_tpr,
        "mean_auc": mean_auc,
        "std_auc": std_auc,
        "std_tpr": std_tpr,
        "mean_f1": mean_f1,
        "std_f1": std_f1,
    }
    return metrics


def get_feature_importance(model):
    if hasattr(model["estimator"], "coef_"):
        return pd.DataFrame(
            model["estimator"].coef_,
            columns=model["estimator"].feature_names_in_,
        )
    elif hasattr(model["estimator"], "feature_importances_"):
        return pd.DataFrame(
            [model["estimator"].feature_importances_],
            columns=model["estimator"].feature_names_in_,
        )
    else:
        print(
            f"The model estimator {model['estimator']} does not support feature importance inference."
        )
        return None



import pytest
from sklearn.datasets import make_classification

import pandas as pd
import numpy as np
from pathlib import Path


@pytest.fixture()
def forest_based_model():
    pass


@pytest.fixture()
def classification_data():
    X, y = make_classification(
        n_samples=100,
        n_features=30,
        n_redundant=10,
        n_classes=2,
    )
    df = pd.DataFrame(X).add_prefix("feature_")
    df.insert(0, "outcome", y)

    # cast to correct dtype and add random np.nan values

    df['outcome'] = df['outcome'].astype(bool)
    df['feature_0'] = df['feature_0'].astype(int)
    df['feature_0'] = df['feature_0'].astype(pd.Int64Dtype())
    df.loc[0, 'feature_0'] = np.nan
    df['feature_1'] = np.random.randint(0, 4, size=df.shape[0])
    df.loc[0, 'feature_1'] = np.nan
    df['feature_1'] = df["feature_1"].astype(pd.CategoricalDtype())
    df.loc[3, 'feature_3'] = np.nan

    # target and feature columns
    target_col = ["outcome"]
    feature_cols = df.columns.tolist()
    feature_cols.remove("outcome")

    return df, target_col, feature_cols


@pytest.fixture()
def output_path():
    return Path("default")



from imblearn.over_sampling import SMOTE, ADASYN, RandomOverSampler
from sklearn.ensemble import (
    RandomForestClassifier,
    AdaBoostClassifier,
    GradientBoostingClassifier,
)
from sklearn.feature_selection import SelectFromModel
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeClassifier
from xgboost import XGBClassifier

classifiers = {
    "logistic_regression": LogisticRegression(n_jobs=-1),
    "random_forest": RandomForestClassifier(n_jobs=-1),
    "ada_boost": AdaBoostClassifier(),
    "gradient_boosting": GradientBoostingClassifier(),
    "decision_tree": DecisionTreeClassifier(),
    "xgboost": XGBClassifier()
    # "k_neighbors": KNeighborsClassifier(n_jobs=-1),
}
sample_methods = {
    "smote": SMOTE(),
    "adasyn": ADASYN(),
    "random_oversampling": RandomOverSampler(),
}
feature_selection_methods = {
    "lasso": SelectFromModel(
        LogisticRegression(penalty="l1", C=0.8, solver="liblinear")
    ),
}
supported_scale_levels = [
    "numerical",
    "ordinal",
    "categorical",
]



import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt


def plot_feature_importances(feature_importances: pd.DataFrame) -> plt.Figure:
    fig = plt.figure(figsize=(9, 7))
    sns.boxplot(data=feature_importances, orient="h", color="cyan", saturation=0.5)
    plt.axvline(x=0, color=".5")
    plt.xlabel("Feature importance")
    plt.title("Feature importance and its variability")
    plt.subplots_adjust(left=0.3)
    return fig


def plot_roc_curve_from_cv_metrics(cv_result_metrics: dict, plot_title: str):
    fig, ax = plt.subplots(figsize=(6, 6))
    ax.plot(
        cv_result_metrics["mean_fpr"],
        cv_result_metrics["mean_tpr"],
        color="b",
        label=r"Mean ROC (AUC = %0.2f $\pm$ %0.2f)"
        % (cv_result_metrics["mean_auc"], cv_result_metrics["std_auc"]),
        lw=2,
        alpha=0.8,
    )

    tprs_upper = np.minimum(
        cv_result_metrics["mean_tpr"] + cv_result_metrics["std_tpr"], 1
    )
    tprs_lower = np.maximum(
        cv_result_metrics["mean_tpr"] - cv_result_metrics["std_tpr"], 0
    )
    ax.fill_between(
        cv_result_metrics["mean_fpr"],
        tprs_lower,
        tprs_upper,
        color="grey",
        alpha=0.2,
        label=r"$\pm$ 1 std. dev.",
    )

    ax.set(
        xlim=[-0.05, 1.05],
        ylim=[-0.05, 1.05],
        xlabel="False Positive Rate",
        ylabel="True Positive Rate",
        title=f"{plot_title}",
    )
    ax.title.set_size(8)
    ax.axis("square")
    ax.legend(loc="lower right")
    return fig



from imblearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import StandardScaler, OneHotEncoder

from racoons.models import classifiers, sample_methods, feature_selection_methods


def get_estimator(estimator_name: str) -> list[tuple[str, object]]:
    """
    Get an estimator as a pipeline step based on its string representation.

    Args:
        estimator_name (str): String representation of the estimator.

    Returns:
        list[tuple[str, object]]: A pipeline-step-like object to be integrated into the model pipeline.

    Raises:
        NotImplementedError: If the specified estimator is not implemented.

    Example:
        >>> get_estimator('random_forest')
        [('estimator', RandomForestClassifier(n_jobs=-1))]

    Note:
        Available classifiers can be found in the 'classifiers' dictionary.
    """
    if estimator_name not in classifiers.keys():
        raise NotImplementedError(
            f"The classifier '{estimator_name}' is not implemented."
        )
    else:
        return [("estimator", classifiers[estimator_name])]


def get_preprocessing_steps(feature_scale_levels: dict) -> list[tuple[str, object]]:
    transformers = []
    if feature_scale_levels["numerical"]:
        transformers.append(
            ("numerical", numerical_preprocessing(), feature_scale_levels["numerical"])
        )
    if feature_scale_levels["ordinal"]:
        transformers.append(
            ("ordinal", ordinal_preprocessing(), feature_scale_levels["ordinal"])
        )
    if feature_scale_levels["categorical"]:
        transformers.append(
            (
                "categorical",
                categorical_preprocessing(),
                feature_scale_levels["categorical"],
            )
        )
    return [("preprocessor", ColumnTransformer(transformers))]


def get_sampling_step(method) -> list[tuple[str, object]]:
    """
    Get a sampling method as a pipeline step.

    Sampling methods are applied to address severe class imbalance in datasets by oversampling the minority class.

    Args:
        method: String representation of the sampler.

    Returns:
        list[tuple[str, object]]: A pipeline-step-like object to be integrated into the model pipeline.

    Raises:
        NotImplementedError: If the specified sampling method is not implemented.

    Example:
        >>> get_sampling_step('smote')
        [('sampling', SMOTE())]

    Note:
        Available sampling methods can be found in the 'sample_methods' dictionary.
    """

    if method is None:
        return []
    elif method not in sample_methods.keys():
        raise NotImplementedError(f"The sampling method '{method}' is not implemented.")
    else:
        return [("sampling", sample_methods[method])]


def get_feature_selection_step(method):
    """
    Get a feature selection method as a pipeline step.

    Feature selection methods are applied to prevent overfitting in datasets with a large number of features
    and a low sample size.

    Args:
        method: String representation of the feature selector.

    Returns:
        list[tuple[str, object]]: A pipeline-step-like object to be integrated into the model pipeline.

    Raises:
        NotImplementedError: If the specified feature selection method is not implemented.

    Example:
        >>> get_feature_selection_step('lasso')
        [('feature_selection', SelectFromModel(LogisticRegression(penalty='l1', C=0.8, solver='liblinear')))]

    Note:
        Available feature selection methods can be found in the 'feature_selection_methods' dictionary.
    """
    if method is None:
        return []
    elif method not in feature_selection_methods.keys():
        raise NotImplementedError(
            f"The feature selection method '{method}' is not implemented."
        )
    else:
        return [("feature_selection", feature_selection_methods[method])]


def numerical_preprocessing():
    pipe = Pipeline(
        [("scaler", StandardScaler()), ("imputer", SimpleImputer(strategy="mean"))]
    )
    return pipe


def ordinal_preprocessing():
    pipe = Pipeline([("imputer", SimpleImputer(strategy="most_frequent"))])
    return pipe


def categorical_preprocessing():
    pipe = Pipeline(
        [
            ("imputer", SimpleImputer(strategy="most_frequent")),
            (
                "encoder",
                OneHotEncoder(
                    sparse_output=False, drop="if_binary", handle_unknown="ignore"
                ),
            ),
        ]
    )
    return pipe


def build_model(
    feature_scale_levels: dict,
    sample_method: str,
    feature_selection_method: str,
    estimator_name: str,
):
    """
    Build a classification model pipeline based on specified components.

    Args:
        feature_scale_levels (dict): Mapping of features to their corresponding scale level.
        sample_method (str): String representation of the sampling method.
        feature_selection_method (str): String representation of the feature selection method.
        estimator_name (str): String representation of the estimator.

    Returns:
        Pipeline: A machine learning model pipeline.

    Raises:
        NotImplementedError: If the specified estimator is not implemented.

    Example:
        >>> build_model({'numerical': ['feature1'], 'ordinal': [], 'categorical': ['feature2']},
        ...             'smote', 'lasso', 'random_forest')
        Pipeline(steps=[('preprocessor', ...), ('sampling', SMOTE()), ('feature_selection', ...), ('estimator', RandomForestClassifier(n_jobs=-1))])

    Note:
        This function assembles a classification model pipeline by combining preprocessing steps,
        sampling, feature selection, and the chosen estimator.
    """
    preprocessor = get_preprocessing_steps(feature_scale_levels)
    sampler = get_sampling_step(sample_method)
    feature_selector = get_feature_selection_step(feature_selection_method)
    estimator = get_estimator(estimator_name)
    return Pipeline(preprocessor + sampler + feature_selector + estimator)



import pandas
import pandas as pd
from pandas.core.dtypes.common import is_bool_dtype, is_float_dtype

from racoons.models import supported_scale_levels


def features_and_targets_from_dataframe(
    df: pd.DataFrame, feature_cols: list[str], target_cols: list[str]
) -> tuple[pd.DataFrame, pd.DataFrame, dict]:
    """
    Split features from targets in a given dataframe and determine the scale level of features.

    Args:
        df (pd.DataFrame): The dataframe containing target columns and feature columns.
        feature_cols (list[str]): Columns containing features.
        target_cols (list[str]): Columns containing targets.

    Returns:
        tuple[pd.DataFrame, pd.DataFrame, dict]: A tuple containing features, targets, and feature scale levels.

    Example:
        >>> df = pd.DataFrame({'feature1': [1, 2, 3], 'target': [0, 1, 0]})
        >>> features, targets, feature_scale_levels = features_and_targets_from_dataframe(df, ['feature1'], ['target'])

    Note:
        This function checks the scale level of feature columns and ensures valid target columns.
    """
    # check scale level of feature columns
    feature_cols_selected = []
    numerical_features, ordinal_features, categorical_features = [], [], []

    for col in feature_cols:
        scale_level = get_scale_level(df[col])
        if scale_level in supported_scale_levels:
            feature_cols_selected.append(col)
            if scale_level == "numerical":
                numerical_features.append(col)
            elif scale_level == "ordinal":
                ordinal_features.append(col)
            elif scale_level == "categorical":
                categorical_features.append(col)
        else:
            print(
                f"Feature column {col} has an unsupported dtype {df[col].dtype} and will be dropped.\n"
            )
    print(
        f"{len(feature_cols_selected)} features out of {len(feature_cols)} "
        f"initial features were selected for analysis.\n"
        f"numerical features: {len(numerical_features)}\n"
        f"ordinal features: {len(ordinal_features)}\n"
        f"categorical features: {len(categorical_features)}\n"
    )
    # check if target columns are binary
    target_cols_selected = []
    for col in target_cols:
        if is_bool_dtype(df[col].dtype):
            target_cols_selected.append(col)
        else:
            print(
                f"Target column {col} has an unsupported dtype {df[col].dtype} and will be dropped.\n"
            )
    print(
        f"{len(target_cols_selected)} targets out of {len(target_cols)} "
        f"initial targets were selected for analysis.\n"
    )

    feature_scale_levels = {
        "numerical": numerical_features,
        "ordinal": ordinal_features,
        "categorical": categorical_features,
    }
    return (
        df.loc[:, feature_cols_selected],
        df.loc[:, target_cols_selected],
        feature_scale_levels,
    )


def get_scale_level(feature: pd.Series) -> str:
    """
    Determine the scale level of a feature based on its data type.

    Args:
        feature (pd.Series): The Pandas Series representing the feature.

    Returns:
        str: The scale level of the feature, which can be one of the following:
            - 'numerical': If the feature has a data type of float.
            - 'ordinal': If the feature has a data type of pd.Int64Dtype.
            - 'categorical': If the feature has a data type of pd.CategoricalDtype.
            - None: If the feature has an unsupported data type.

    Raises:
        None

    Example:
        >>> import pandas as pd
        >>> feature = pd.Series([1, 2, 3], dtype=float)
        >>> get_scale_level(feature)
        'numerical'

    Note:
        This function is designed to determine the scale level of a feature based on its data type.
        If the data type is not supported, it returns None.
    """
    if is_float_dtype(feature.dtype):
        return "numerical"
    elif isinstance(feature.dtype, pd.Int64Dtype):
        return "ordinal"
    elif isinstance(feature.dtype, pd.CategoricalDtype):
        return "categorical"
    else:
        print(
            f"The feature {feature.name} has an unsupported dtype '{feature.dtype}' and will be dropped."
        )



import pytest
import pandas as pd
from racoons.data_utils import features_and_targets_from_dataframe, get_scale_level


@pytest.mark.parametrize(
    "test_input, expected_output",
    [
        (pd.Series([1, 2, 3], dtype=float), "numerical"),
        (pd.Series([1, 2, 3], dtype=pd.Int64Dtype()), "ordinal"),
        (pd.Series(["a", "b", "c"], dtype="category"), "categorical"),
        (pd.Series([1, 2, 3], dtype=int), None),  # Unsupported dtype, expecting None
        (pd.Series([True, False, True], dtype=bool), None),  # Unsupported dtype, expecting None
    ],
)
def test_get_scale_level(test_input, expected_output):
    result = get_scale_level(test_input)
    assert result == expected_output


def test_features_and_targets_from_dataframe(classification_data):
    df, target_cols, feature_cols = classification_data
    X, y, feature_scale_levels = features_and_targets_from_dataframe(
        df,
        feature_cols=feature_cols,
        target_cols=target_cols,
    )

    assert y.columns.tolist() == target_cols
    assert X.columns.tolist() == feature_cols
    assert feature_scale_levels["ordinal"] == ["feature_0"]
    assert feature_scale_levels["categorical"] == ["feature_1"]
    assert feature_scale_levels["numerical"] == [f"feature_{i}" for i in range(2, len(feature_cols))]





import pandas as pd


def make_report_df(sampling: str, feature_selection: str, grid_search=False):
    """
    Creates an empty dataframe to store the results of the classification task.
    Args:
        sampling (str): sample method string representation
        feature_selection (str): feature selection string representation
        grid_search (bool): grid search optimization was conducted

    Returns:
        pd.DataFrame: empty dataframe to store the classification results
    """
    columns = [
        "Outcome",
        "Class distribution",
        "Covariates",
        "Model",
        "AUC_mean",
        "AUC_std",
        "f1_mean",
        "f1_std",
        "feature_importance",
        "roc_curve_plot_path",
        "feature_importance_plot_path",
    ]

    if feature_selection is not None:
        columns.insert(columns.index("Model"), "Number of selected covariates")
        columns.insert(
            columns.index("Number of selected covariates") + 1,
            "Feature selection method",
        )

    if sampling is not None:
        columns.insert(columns.index("Model"), "Sampling method")

    if grid_search:
        columns.insert(columns.index("Model"), "Best_params")
    return pd.DataFrame(columns=columns)


def update_report(
    target,
    features,
    negative_samples,
    positive_samples,
    estimator_name,
    mean_auc,
    std_auc,
    mean_f1,
    std_f1,
    feature_importance_csv,
    roc_plot_path,
    feature_importance_plot_path,
    sampling=None,
    feature_selection=None,
    selected_features=None,
    best_params=None,
):
    row = [
        target,
        f"positive: {positive_samples}, negative: {negative_samples}",
        features,
        estimator_name,
        mean_auc,
        std_auc,
        mean_f1,
        std_f1,
        feature_importance_csv,
        roc_plot_path,
        feature_importance_plot_path,
    ]
    if feature_selection is not None:
        row.insert(row.index(estimator_name), feature_selection)
        row.insert(row.index(feature_selection), len(selected_features))
    if sampling is not None:
        row.insert(row.index(estimator_name), sampling)
    if best_params:
        row.insert(row.index(estimator_name), f"{best_params}")
    return row



from datetime import datetime

import pandas as pd
from matplotlib import pyplot as plt
import sklearn
from pathlib import Path
from tqdm import tqdm

from racoons.models import classifiers
from racoons.data_utils import features_and_targets_from_dataframe
from racoons.models.model_builder import get_estimator, build_model
from racoons.reporting import make_report_df, update_report
from racoons.models.validation import (
    hyper_parameter_optimization,
    cross_validate_model,
    metrics_from_cv_result,
    get_feature_importance,
)
from racoons.visualization import (
    plot_feature_importances,
    plot_roc_curve_from_cv_metrics,
)

######################## OPEN ISSUES ##################################################
# TODO: refit on f1 in cross-validation
#######################################################################################

sklearn.set_config(transform_output="pandas")


def multivariate_classification(
    df: pd.DataFrame,
    feature_cols: list[str],
    target_cols: list[str],
    feature_selection_method: str,
    sample_method: str,
    estimators: list[str],
    output_path: Path,
):
    """
    Run a classification pipeline with cross-validation and save the results.

    Args:
        df (pd.DataFrame): The input DataFrame containing features and targets.
        feature_cols (List[str]): List of column names containing features.
        target_cols (List[str]): List of column names containing target variables.
        feature_selection_method (Optional[str]): The method for feature selection, if any.
        sample_method (Optional[str]): The method for addressing class imbalance, if any.
        estimators (List[str]): List of classifier names to use in the pipeline.
        output_path (Optional[Path]): Path to the output folder. Defaults to 'output'.
        **kwargs: Additional keyword arguments.

    Returns:
        pd.DataFrame: A DataFrame containing the results of the classification.

    Raises:
        Any specific exceptions or errors.

    Example:
        >>> result_df = multivariate_classification(
        ...     df=data_frame,
        ...     feature_cols=['feature1', 'feature2'],
        ...     target_cols=['target'],
        ...     feature_selection_method='lasso',
        ...     sample_method='smote',
        ...     estimators=['random_forest', 'logistic_regression'],
        ...     output_path=Path('results'),
        ... )

    Note:
        - The function runs a classification pipeline for each target column and each specified estimator.
        - Cross-validation is performed, and results including ROC-AUC and feature importance are saved.
        - The output is stored in the specified output_path.
    """
    # setup output folder
    output_path.mkdir(exist_ok=True)
    output_folder = (
        output_path
        / f"multivariate_analysis_{datetime.now().strftime('%Y-%m-%d_%H:%M:%S')}"
    )
    output_folder.mkdir(exist_ok=True)

    # get features and targets
    features, targets, feature_scale_levels = features_and_targets_from_dataframe(
        df, feature_cols, target_cols
    )

    # iterate over targets and features
    report_df = make_report_df(sample_method, feature_selection_method)
    with tqdm(total=(targets.shape[1]) * len(classifiers)) as pbar:
        plot_index = 0
        for target in targets:
            for estimator in estimators:
                estimator_name = type(get_estimator(estimator)[0][1]).__name__
                model = build_model(
                    feature_scale_levels,
                    sample_method,
                    feature_selection_method,
                    estimator,
                )
                cv_result = cross_validate_model(model, df[feature_cols], df[target])
                cv_result_metrics = metrics_from_cv_result(cv_result)

                # roc-auc
                roc_curve_plot = plot_roc_curve_from_cv_metrics(
                    cv_result_metrics,
                    plot_title=f"Classification of {target} using " f"{estimator_name}",
                )
                roc_curve_plot_path = output_folder / (f"roc_auc_{plot_index}.png")
                roc_curve_plot.savefig(roc_curve_plot_path, dpi=600)
                plt.close(roc_curve_plot)

                # feature importance
                feature_importance = get_feature_importance(model)
                feature_importance_plot = plot_feature_importances(feature_importance)
                feature_importance_plot_path = output_folder / (
                    f"feature_importance_{plot_index}.png"
                )
                feature_importance_csv_path = output_folder / (
                    f"feature_importance_{plot_index}.csv"
                )
                feature_importance.to_csv(feature_importance_csv_path, sep=";")
                feature_importance_plot.savefig(feature_importance_plot_path, dpi=300)
                plt.close(feature_importance_plot)

                # save report
                selected_features = model["estimator"].feature_names_in_
                negative_samples = (~df[target]).sum()
                positive_samples = (df[target]).sum()
                mean_auc = cv_result_metrics["mean_auc"]
                std_auc = cv_result_metrics["std_auc"]
                mean_f1 = cv_result_metrics["mean_f1"]
                std_f1 = cv_result_metrics["std_f1"]

                report_df.loc[len(report_df.index)] = update_report(
                    target=target,
                    features=feature_cols,
                    negative_samples=negative_samples,
                    positive_samples=positive_samples,
                    estimator_name=estimator_name,
                    mean_auc=mean_auc,
                    std_auc=std_auc,
                    mean_f1=mean_f1,
                    std_f1=std_f1,
                    roc_plot_path=roc_curve_plot_path,
                    feature_importance_csv=feature_importance_csv_path,
                    feature_importance_plot_path=feature_importance_plot_path,
                    sampling=sample_method,
                    feature_selection=feature_selection_method,
                    selected_features=selected_features,
                )
                pbar.update(1)
                plot_index += 1
            report_df.to_excel(output_folder / "report.xlsx")
            report_df.to_csv(output_folder / "report.csv", sep=";")

        report_df.to_excel(output_folder / "report.xlsx")
        report_df.to_csv(output_folder / "report.csv", sep=";")
        return report_df


def grid_search_multivariate_classification(
    df: pd.DataFrame,
    feature_cols: list[str],
    target_cols: list[str],
    feature_selection_method: str,
    sample_method: str,
    estimators: list[str],
    output_path: Path,
):
    """
    Perform grid search for hyperparameter optimization in a classification pipeline.

    This function conducts a grid search to find the optimal hyperparameters for each specified classifier
    while considering feature selection and sampling methods. The grid search is performed based on cross-validation
    metrics (AUC and F1 score) to evaluate the model's performance.

    Args:
        df (pd.DataFrame): The input dataframe containing features and targets.
        feature_cols (List[str]): List of column names representing the feature variables.
        target_cols (List[str]): List of column names representing the target variables.
        feature_selection_method (str): The method used for feature selection. Supported methods are 'lasso' and None.
        sample_method (str): The method used for handling class imbalance. Supported methods are 'smote', 'adasyn', and 'random_oversampling'.
        estimators (List[str]): List of classifiers to be evaluated during the grid search.
        output_path (Path): The output path where the results and plots will be saved.

    Returns:
        pd.DataFrame: A dataframe containing the results of the grid search, including key metrics and paths to relevant plots.

    Raises:
        None

    Example:
        >>> df = pd.DataFrame({'feature1': [1, 2, 3], 'target': [0, 1, 0]})
        >>> feature_cols = ['feature1']
        >>> target_cols = ['target']
        >>> feature_selection_method = 'lasso'
        >>> sample_method = 'smote'
        >>> estimators = ['random_forest', 'gradient_boosting']
        >>> output_path = Path('./output/')
        >>> result_df = grid_search_multivariate_classification(df, feature_cols, target_cols,
        ...                                       feature_selection_method, sample_method,
        ...                                       estimators, output_path)

    Note:
        This function performs grid search by evaluating the specified classifiers with various hyperparameters,
        considering the provided feature selection and sampling methods. It outputs a dataframe with the grid search results.
    """
    # setup output folder
    output_path.mkdir(exist_ok=True)
    output_folder = (
        output_path
        / f"gs_multivariate_analysis_{datetime.now().strftime('%Y-%m-%d_%H:%M:%S')}"
    )
    output_folder.mkdir(exist_ok=True)

    # get features and targets
    features, targets, feature_scale_levels = features_and_targets_from_dataframe(
        df, feature_cols, target_cols
    )

    # iterate over targets and features
    report_df = make_report_df(
        sample_method, feature_selection_method, grid_search=True
    )
    with tqdm(total=(targets.shape[1])) as pbar:
        # find best performing model by cross validation
        plot_index = 0
        for target in targets:
            cv_df = pd.DataFrame(columns=["Model", "AUC"])
            print("Initial cross-validation...")
            for estimator in estimators:
                # remove feature selection steps from model for initial cv
                model = build_model(
                    feature_scale_levels, sample_method, None, estimator
                )
                initial_cv_result = cross_validate_model(
                    model, df[feature_cols], df[target]
                )

                cv_result_metrics = metrics_from_cv_result(initial_cv_result)
                cv_df.loc[len(cv_df.index)] = [
                    estimator,
                    cv_result_metrics["mean_auc"],
                ]
            print("Done.")
            best_cv_estimator = cv_df.sort_values(by="AUC", ascending=False)[
                "Model"
            ].iloc[0]
            grid_search_model = build_model(
                feature_scale_levels,
                sample_method,
                feature_selection_method,
                best_cv_estimator,
            )
            print("Hyper-parameter optimization...")
            best_params, mean_score, std_score = hyper_parameter_optimization(
                grid_search_model,
                df[feature_cols],
                df[target],
            )
            print("done.")
            # run cross valiadtion with best model
            optimized_model = build_model(
                feature_scale_levels,
                sample_method,
                feature_selection_method,
                best_cv_estimator,
            )
            optimized_model.set_params(**best_params)
            cv_result = cross_validate_model(
                optimized_model, df[feature_cols], df[target]
            )
            cv_result_metrics = metrics_from_cv_result(cv_result)

            # roc-auc
            roc_curve_plot = plot_roc_curve_from_cv_metrics(
                cv_result_metrics,
                plot_title=f"Classification of {target} using " f"{best_cv_estimator}",
            )
            roc_curve_plot_path = output_folder / (f"roc_auc_{plot_index}.png")
            roc_curve_plot.savefig(roc_curve_plot_path, dpi=300)
            plt.close(roc_curve_plot)

            # feature importance
            feature_importance = get_feature_importance(model)
            feature_importance_plot = plot_feature_importances(feature_importance)
            feature_importance_plot_path = output_folder / (
                f"feature_importance_{plot_index}.png"
            )
            feature_importance_csv_path = output_folder / (
                f"feature_importance_{plot_index}.csv"
            )
            feature_importance.to_csv(feature_importance_csv_path, sep=";")
            feature_importance_plot.savefig(feature_importance_plot_path, dpi=300)
            plt.close(feature_importance_plot)

            # save report
            selected_features = optimized_model["estimator"].feature_names_in_.tolist()
            negative_samples = (~df[target]).sum()
            positive_samples = (df[target]).sum()
            mean_auc = cv_result_metrics["mean_auc"]
            std_auc = cv_result_metrics["std_auc"]
            mean_f1 = cv_result_metrics["mean_f1"]
            std_f1 = cv_result_metrics["std_f1"]

            report_df.loc[len(report_df.index)] = update_report(
                target=target,
                features=feature_cols,
                negative_samples=negative_samples,
                positive_samples=positive_samples,
                estimator_name=type(classifiers[best_cv_estimator]).__name__,
                mean_auc=mean_auc,
                std_auc=std_auc,
                mean_f1=mean_f1,
                std_f1=std_f1,
                roc_plot_path=roc_curve_plot_path,
                feature_importance_csv=feature_importance_csv_path,
                feature_importance_plot_path=feature_importance_plot_path,
                sampling=sample_method,
                best_params=best_params,
                feature_selection=feature_selection_method,
                selected_features=selected_features,
            )
            pbar.update(1)
            plot_index += 1
            report_df.to_excel(output_folder / "report.xlsx")
            report_df.to_csv(output_folder / "report.csv", sep=";")
        report_df.to_excel(output_folder / "report.xlsx")
        report_df.to_csv(output_folder / "report.csv", sep=";")
        return report_df


def univariate_classification(
    df: pd.DataFrame,
    feature_cols: list[str],
    target_cols: list[str],
    sample_method: str,
    estimators: list[str],
    output_path: Path,
):
    """
    Run a univariate classification pipeline with cross-validation and save the results.

    Args:
        df (pd.DataFrame): The input DataFrame containing features and targets.
        feature_cols (List[str]): List of column names containing features.
        target_cols (List[str]): List of column names containing target variables.
        sample_method (Optional[str]): The method for addressing class imbalance, if any.
        estimators (List[str]): List of classifier names to use in the pipeline.
        output_path (Optional[Path]): Path to the output folder. Defaults to 'output'.

    Returns:
        pd.DataFrame: A DataFrame containing the results of the classification.

    Raises:
        Any specific exceptions or errors.

    Example:
        >>> result_df = multivariate_classification(
        ...     df=data_frame,
        ...     feature_cols=['feature1', 'feature2'],
        ...     target_cols=['target'],
        ...     sample_method='smote',
        ...     estimators=['random_forest', 'logistic_regression'],
        ...     output_path=Path('results'),
        ... )

    Note:
        - The function runs a univariate classification pipeline for each target column and each specified estimator.
        - Cross-validation is performed, and results including ROC-AUC and feature importance are saved.
        - The output is stored in the specified output_path.
    """
    feature_selection_method = None  # redundant for univariate analysis
    with tqdm(total=(len(target_cols) * len(feature_cols) * len(classifiers))) as pbar:
        # setup output folder
        output_path.mkdir(exist_ok=True)
        output_folder = (
            output_path
            / f"univariate_analysis_{datetime.now().strftime('%Y-%m-%d_%H:%M:%S')}"
        )
        output_folder.mkdir(exist_ok=True)

        report_df = make_report_df(sample_method, feature_selection_method)
        # iterate over feaures
        plot_index = 0
        for feature in feature_cols:
            feature_cols_ = [feature]

            # get features and targets
            (
                features,
                targets,
                feature_scale_levels,
            ) = features_and_targets_from_dataframe(df, feature_cols_, target_cols)

            # iterate over targets

            for target in targets:
                for estimator in estimators:
                    estimator_name = type(classifiers[estimator]).__name__
                    model = build_model(
                        feature_scale_levels,
                        sample_method,
                        feature_selection_method,
                        estimator,
                    )
                    cv_result = cross_validate_model(
                        model, df[feature_cols_], df[target]
                    )
                    cv_result_metrics = metrics_from_cv_result(cv_result)

                    # roc-auc
                    roc_curve_plot = plot_roc_curve_from_cv_metrics(
                        cv_result_metrics,
                        plot_title=f"Classification of {target} using "
                        f"{estimator_name}",
                    )
                    roc_curve_plot_path = output_folder / (f"roc_auc_{plot_index}.png")
                    roc_curve_plot.savefig(roc_curve_plot_path, dpi=600)
                    plt.close(roc_curve_plot)

                    # feature importance
                    feature_importance = get_feature_importance(model)
                    feature_importance_plot = plot_feature_importances(
                        feature_importance
                    )
                    feature_importance_plot_path = output_folder / (
                        f"feature_importance_{plot_index}.png"
                    )
                    feature_importance_csv_path = output_folder / (
                        f"feature_importance_{plot_index}.csv"
                    )
                    feature_importance.to_csv(feature_importance_csv_path, sep=";")
                    feature_importance_plot.savefig(
                        feature_importance_plot_path, dpi=300
                    )
                    plt.close(feature_importance_plot)

                    # save report
                    features = model["estimator"].feature_names_in_.tolist()
                    negative_samples = (~df[target]).sum()
                    positive_samples = (df[target]).sum()
                    mean_auc = cv_result_metrics["mean_auc"]
                    std_auc = cv_result_metrics["std_auc"]
                    mean_f1 = cv_result_metrics["mean_f1"]
                    std_f1 = cv_result_metrics["std_f1"]

                    report_df.loc[len(report_df.index)] = update_report(
                        target=target,
                        features=features,
                        negative_samples=negative_samples,
                        positive_samples=positive_samples,
                        estimator_name=estimator_name,
                        mean_auc=mean_auc,
                        std_auc=std_auc,
                        mean_f1=mean_f1,
                        std_f1=std_f1,
                        roc_plot_path=roc_curve_plot_path,
                        feature_importance_csv=feature_importance_csv_path,
                        feature_importance_plot_path=feature_importance_plot_path,
                        sampling=sample_method,
                    )
                    pbar.update(1)
                    plot_index += 1
                report_df.to_excel(output_folder / "report.xlsx")
                report_df.to_csv(output_folder / "report.csv", sep=";")

            report_df.to_excel(output_folder / "report.xlsx")
            report_df.to_csv(output_folder / "report.csv", sep=";")
        return report_df



import pytest
from imblearn.pipeline import Pipeline

from racoons.data_utils import features_and_targets_from_dataframe
from racoons.models import classifiers, sample_methods, feature_selection_methods
from racoons.models.model_builder import get_estimator, get_preprocessing_steps, get_sampling_step, \
    get_feature_selection_step, build_model


def test_get_estimator():
    for classifier in classifiers.keys():
        estimator = get_estimator(classifier)
        assert isinstance(estimator[0][1], type(classifiers[classifier]))
    with pytest.raises(NotImplementedError):
        get_estimator("unknown_estimator")


def test_get_preprocessing_steps(classification_data):
    df, target_cols, feature_cols = classification_data
    X, y, scale_levels = features_and_targets_from_dataframe(df, feature_cols, target_cols)
    preprocessing_steps = get_preprocessing_steps(scale_levels)
    transformers = preprocessing_steps[0][1].transformers
    assert transformers[0][0] == "numerical"
    assert transformers[0][2] == [f"feature_{i}" for i in range(2, len(feature_cols))]
    assert transformers[1][0] == "ordinal"
    assert transformers[1][2] == ["feature_0"]
    assert transformers[2][0] == "categorical"
    assert transformers[2][2] == ["feature_1"]


def test_get_sampling_steps():
    for sample_method in sample_methods.keys():
        sampler = get_sampling_step(sample_method)
        assert isinstance(sampler[0][1], type(sample_methods[sample_method]))
    with pytest.raises(NotImplementedError):
        get_sampling_step("unknown_sample_method")


def test_get_feature_selection_steps():
    for feature_selection_method in feature_selection_methods.keys():
        selector = get_feature_selection_step(feature_selection_method)
        assert isinstance(selector[0][1], type(feature_selection_methods[feature_selection_method]))
    with pytest.raises(NotImplementedError):
        get_feature_selection_step("unknown_sample_method")


def test_build_model(classification_data):
    df, target_cols, feature_cols = classification_data
    X, y, scale_levels = features_and_targets_from_dataframe(df, feature_cols, target_cols)

    sample_method = "smote"
    feature_selection_method = "lasso"
    estimator_name = "random_forest"

    model = build_model(scale_levels, sample_method, feature_selection_method, estimator_name)

    assert isinstance(model, Pipeline)
    assert "preprocessor" in model.named_steps
    assert "sampling" in model.named_steps
    assert "feature_selection" in model.named_steps
    assert "estimator" in model.named_steps
    assert model["sampling"].__class__.__name__ == "SMOTE"
    assert model["feature_selection"].__class__.__name__ == "SelectFromModel"
    assert model["estimator"].__class__.__name__ == "RandomForestClassifier"
    assert model["estimator"].n_jobs == -1



import numpy as np

from racoons.models import classifiers
from racoons.models.validation import get_param_grid, cross_validate_model, get_feature_importance
from racoons.models.model_builder import build_model
from racoons.models.classification import multivariate_classification, grid_search_multivariate_classification, univariate_classification
from racoons.visualization import plot_feature_importances
from racoons.data_utils import features_and_targets_from_dataframe


def test_feature_importance(classification_data):
    df, target_cols, feature_cols = classification_data
    X, y, scale_levels = features_and_targets_from_dataframe(df, feature_cols, target_cols)

    sample_method = "smote"
    feature_selection_method = "lasso"
    estimator_name = "logistic_regression"

    model = build_model(scale_levels, sample_method, feature_selection_method, estimator_name)
    model.fit(X, y)
    feature_importance = get_feature_importance(model)
    plot_feature_importances(feature_importance)
    assert feature_importance.columns.tolist() == model["estimator"].feature_names_in_.tolist()
    assert not feature_importance.empty


class TestClassification:
    def test_cross_validate_model(self, classification_data):
        df, target_cols, feature_cols = classification_data
        X, y, scale_levels = features_and_targets_from_dataframe(df, feature_cols, target_cols)

        sample_method = "smote"
        feature_selection_method = "lasso"
        estimator_name = "random_forest"

        model = build_model(scale_levels, sample_method, feature_selection_method, estimator_name)

        # Test the cross_validate_model function
        tprs, aucs, f1_scores, feature_importances = cross_validate_model(model, X, y["outcome"])


        # Check if the output lists have the correct lengths (10 folds)
        assert len(tprs) == len(aucs) == len(f1_scores) == 10

        # Check if feature importances is not empty
        assert not feature_importances.empty

        # Check if each element in the lists is a NumPy array
        assert all(isinstance(tpr, np.ndarray) for tpr in tprs)
        assert all(isinstance(auc_, np.float64) for auc_ in aucs)
        assert all(isinstance(f1_score_, np.float64) for f1_score_ in f1_scores)

    def test_multivariate_classification(self, classification_data, output_path, tmp_path):
        out_path = tmp_path
        df, target_cols, feature_cols = classification_data

        result_df = multivariate_classification(
            df=df,
            feature_cols=feature_cols,
            target_cols=target_cols,
            feature_selection_method="lasso",
            sample_method="smote",
            estimators=classifiers.keys(),
            output_path=out_path,
        )

        # Validate the result
        assert len(result_df) == len(target_cols) * len(classifiers)
        assert not result_df.empty

    def test_univariate_classification(self, classification_data, tmp_path, output_path):
        out_path = tmp_path
        df, target_cols, feature_cols = classification_data

        result_df = univariate_classification(
            df=df,
            feature_cols=feature_cols,
            target_cols=target_cols,
            sample_method="smote",
            estimators=classifiers.keys(),
            output_path=out_path,
        )

        # Validate the result
        assert len(result_df) == len(target_cols) * len(classifiers) * len(feature_cols)
        assert not result_df.empty


class TestGridSearchClassification:
    def test_get_param_grid(self, classification_data):
        for classifier_name, classifier in classifiers.items():
            df, target_cols, feature_cols = classification_data
            X, y, scale_levels = features_and_targets_from_dataframe(df, feature_cols, target_cols)
            model = build_model(
                feature_scale_levels=scale_levels,
                feature_selection_method=None,
                sample_method="smote",
                estimator_name=classifier_name
            )
            param_grid = get_param_grid(model)

            if classifier_name == "logistic_regression":
                assert "estimator__penalty" in param_grid
                assert "estimator__solver" in param_grid
                assert "estimator__C" in param_grid
            elif classifier_name == "random_forest":
                assert "estimator__n_estimators" in param_grid
                assert "estimator__n_jobs" in param_grid
            elif classifier_name == "ada_boost":
                assert "estimator__learning_rate" in param_grid
                assert "estimator__n_estimators" in param_grid
            elif classifier_name == "gradient_boosting":
                assert "estimator__learning_rate" in param_grid
                assert "estimator__n_estimators" in param_grid
            elif classifier_name == "decision_tree":
                assert "estimator__criterion" in param_grid
            elif classifier_name == "xgboost":
                assert "estimator__n_estimators" in param_grid
            # elif classifier_name == "k_neighbors":
            #     assert "estimator__n_neighbors" in param_grid
            #     assert "estimator__n_jobs" in param_grid

            model = build_model(
                feature_scale_levels=scale_levels,
                feature_selection_method="lasso",
                sample_method="smote",
                estimator_name=classifier_name
            )
            param_grid = get_param_grid(model)
            assert "feature_selection__estimator__C" in param_grid
            assert "feature_selection__estimator__solver" in param_grid

    def test_grid_search_classification(self, classification_data, tmp_path, output_path):
        df, target_cols, feature_cols = classification_data
        out_path = tmp_path
        result_df = grid_search_multivariate_classification(
            df=df,
            feature_cols=feature_cols,
            target_cols=target_cols,
            feature_selection_method="lasso",
            sample_method="smote",
            estimators=classifiers.keys(),
            output_path=out_path,
        )

        # Validate the result
        assert len(result_df) == len(target_cols)
        assert not result_df.empty


